/**************************************
 * TGame.cpp
 * Copyright  1996 Steve Sprang
 * All Rights Reserved 
 **************************************/

#include "TGame.h"
#include "Views.h"
#include "Tetris.h"
#include "Board.h"
#include "TWindow.h"
#include "BeTetris.h"
#include "THighScore.h"

TGame::TGame(TWindow *myWindow, int start_level, int debris_rows)
{
	window_pointers_t 	v;
	
	w = myWindow;
	
	v = w->GetViewPointers();
	
	arena = v.arena;
	arenaview = v.arenaview;
	arenaoff = v.arenaoff;
	nextview = v.nextview;
	nextBlit = v.next;
	scoreview = v.score;
	linesview = v.lines;
	levelview = v.level;
	high_scores = v.hs;
	
	next = current = NULL;
	score = lines = 0;
	start = level = start_level;
	pause = done = removing_rows = FALSE;
	
	UpdateScore();
	theBoard = new Board(arenaview, arena, arenaoff, debris_rows);
	
	timer_thread = spawn_thread(TGame::Timer, "", B_NORMAL_PRIORITY, this);
	resume_thread(timer_thread);
}

void TGame::SetKeys(ulong l, ulong r, ulong c, ulong cc, ulong d)
{
	left = l;
	right = r;
	cw = c;
	ccw = cc;
	drop = d;
}

bool TGame::PausedQ()
{
	return(pause);
}

long TGame::Timer(void *arg)
{
	return ((TGame *) arg)->TimerLoop();
}

long TGame::TimerLoop()
{
	int	time;
	
	while(!done) {
		time = (550 - (level * 50)) * 1000;
		if(time <= 0) {
			time = 25000;
		}
		snooze(time);
		if(!removing_rows) {
			PostMessage(new BMessage(mPULSE));
		}
	}
	return(find_thread(NULL));
}

void TGame::TogglePause()
{
	pause = !pause;
	
	if(pause) {
		ShowMessage("PAUSED", FALSE);
	}
	else {
		theBoard->Draw();
		current->Draw();
	}
}

void TGame::ShiftPieceDown()
{
	int		stop;
	
	if(next == NULL && current == NULL) {
		current = new Tetris(theBoard, arenaview, this);
		next = new Tetris(theBoard, arenaview, this);
		DrawNext();
	}
	else {
		stop = current->ShiftDown();
		if(stop == 1)	
			Stop();
		else if(stop > 1)
			GameOver();
	}
}

void TGame::GameOver()
{
	score_t s;
	
	w->SetGameToNull();
	
	done = TRUE;
	kill_thread(timer_thread);
	
	delete theBoard;
	if(current) delete current;
	delete next;
	
	ShowMessage("GAME OVER", TRUE);

	ClearNext();
	
	s.score = score;
	s.lines = lines;
	s.level = level;
	high_scores->InsertNewScore(s);
	
	// tell the window we're done
	
	Quit();
}

void TGame::ClearNext()
{
	nextview->Window()->Lock();
	nextview->SetDrawingMode(B_OP_COPY);
	nextview->SetHighColor(0,0,0);
	nextview->FillRect(nextview->Bounds());
	nextview->Sync();
	nextview->Window()->Unlock();
	nextBlit->Draw();
}

void TGame::MyQuit()
{
	done = TRUE;
	
	delete theBoard;
	delete current;
	delete next;
	
	Quit();
}

void TGame::Stop()
{
	int num_rows;
	
	// set this flag to keep TimerLoop() from pulsing
	removing_rows = TRUE;
	
	// make next the current and pick the new next
	delete current;
	current = next;
	next = new Tetris(theBoard, arenaview, this);
	
	// count and remove any completed rows
	num_rows = theBoard->RemoveCompletedRows();
	
	// update stats
	lines += num_rows;
	score += (num_rows * num_rows * 10 * level) + (level * 5);
	
	UpdateScore();
	if(lines && lines >= 10 * (level-start+1)) {
		beep();
		level++;
	}
	
	// show the new next
	DrawNext();
	
	// let TimerLoop() pulse again
	removing_rows = FALSE;
}

void TGame::DrawNext()
{
	next->DrawNext(nextview, nextBlit);
}

void TGame::UpdateScore()
{
	w->Lock();
	scoreview->SetValue(score);
	linesview->SetValue(lines);
	levelview->SetValue(level);
	w->Unlock();
}

void TGame::ShowMessage(const char *message, bool clear)
{
	BRect		bounds;
	int 		length, h_offset, v_offset;
	int 		size = 24;
	
	arenaview->Window()->Lock();
	
	arenaview->SetDrawingMode(B_OP_COPY);
	bounds = arenaview->Bounds();
	
	if(clear) {
		arenaview->SetHighColor(0,0,0);
		arenaview->FillRect(bounds);
	}
	
	arenaview->SetFontName("Arial MT Bold");
	arenaview->SetFontSize(size);
	
	length = arenaview->StringWidth(message);
	h_offset = (bounds.Width() - length) / 2;
	v_offset = (bounds.Height() - size) / 2 + (size * .5);
	
	arenaview->SetHighColor(92,92,92);
	arenaview->DrawString(message, BPoint(h_offset+1, v_offset+2));
	arenaview->SetHighColor(255,255,255);
	arenaview->DrawString(message, BPoint(h_offset, v_offset));
	arenaview->Sync();
	arenaview->Window()->Unlock();
	
	arena->Draw();
}

void TGame::HandleKeys(ulong aChar)
{
	if(aChar == drop) {
		if(current->Drop()) {
			Stop();
		}
	}
	else if(aChar == left) {
		current->Shift(LEFT);
	}
	else if(aChar == right) {
		current->Shift(RIGHT);
	}
	else if(aChar == ccw) {
		current->Rotate(CCW);
	}
	else if(aChar == cw) {
		current->Rotate(CW);
	}
}

void TGame::MessageReceived(BMessage *msg)
{
	Lock();
	switch(msg->what) {
		case mPULSE:
			if(!pause) {
				ShiftPieceDown();
			}
		break;
			
		case mQUIT:
			MyQuit();
		break;
			
		case mKEYDOWN:
			if(current != NULL && !pause) {
				HandleKeys(msg->FindLong("key"));
			}
		break;
		}
	Unlock();
}